import typing
from abc import ABC
from abc import abstractmethod

import httpx
from httpx._client import USE_CLIENT_DEFAULT
from httpx._client import Response
from httpx._types import CertTypes
from httpx._types import CookieTypes
from httpx._types import ProxiesTypes
from httpx._types import TimeoutTypes
from httpx._types import URLTypes
from httpx._types import VerifyTypes

from scrapyio.utils import first_not_none

from . import default_configs
from .exceptions import IgnoreRequestError
from .http import Request
from .http import clean_up_response
from .middlewares import BaseMiddleWare
from .middlewares import build_middlewares_chain
from .types import CLEANUP_WITH_RESPONSE


class BaseDownloader(ABC):
    def __init__(self):
        self.middlewares: typing.List[BaseMiddleWare] = build_middlewares_chain()

    async def _send_request_via_middlewares(
        self, request: "Request"
    ) -> typing.Union[None, CLEANUP_WITH_RESPONSE]:
        for middleware in self.middlewares:
            resp = await middleware.process_request(request=request)
            if resp is not None:
                if isinstance(resp, tuple):
                    return resp
                else:
                    raise TypeError(
                        "Request processing middleware must return "
                        "either `Tuple[CLEANUP_WITH_RESPONSE]` or "
                        "`None` not `%s`" % resp.__class__.__name__
                    )

    async def _send_response_via_middlewares(
        self, response: Response
    ) -> typing.Union[None, Request]:
        for middleware in reversed(self.middlewares):
            request = await middleware.process_response(response=response)
            if request is not None:
                if isinstance(request, Request):
                    return request
                else:
                    raise TypeError(
                        "Response processing middleware must return "
                        "either `Request` or `None` not `%s`"
                        % request.__class__.__name__
                    )

    def send_request(self, request: "Request") -> typing.AsyncGenerator[Response, None]:
        return send_request(request=request)

    async def _process_request_with_middlewares(
        self, request: "Request"
    ) -> typing.Optional[CLEANUP_WITH_RESPONSE]:
        try:
            cleanup_and_response = await self._send_request_via_middlewares(
                request=request
            )
            if cleanup_and_response is None:
                clean_up = self.send_request(request=request)
                response = await clean_up.__anext__()
            else:
                clean_up, response = cleanup_and_response
            next_request = await self._send_response_via_middlewares(response=response)
            if next_request is not None:
                await clean_up_response(clean_up)
                return await self._process_request_with_middlewares(
                    request=next_request
                )
            return clean_up, response
        except IgnoreRequestError:
            return None

    @abstractmethod
    async def handle_request(
        self, request: "Request"
    ) -> typing.Optional[CLEANUP_WITH_RESPONSE]:
        ...


class Downloader(BaseDownloader):
    async def handle_request(
        self, request: "Request"
    ) -> typing.Optional[CLEANUP_WITH_RESPONSE]:
        response_gen = await self._process_request_with_middlewares(request=request)
        return response_gen


class SessionDownloader(BaseDownloader):
    def __init__(
        self,
        app: typing.Optional[typing.Callable[..., typing.Any]] = None,
        base_url: URLTypes = "",
        cookies: typing.Optional[CookieTypes] = None,
        proxies: typing.Optional[ProxiesTypes] = None,
        verify: typing.Optional[VerifyTypes] = None,
        cert: typing.Optional[CertTypes] = None,
        http1: typing.Optional[bool] = None,
        http2: typing.Optional[bool] = None,
        timeout: typing.Optional[TimeoutTypes] = None,
        trust_env: typing.Optional[bool] = None,
    ):
        super().__init__()
        self.session: httpx.AsyncClient = create_default_session(
            app=app,
            base_url=base_url,
            cookies=cookies,
            proxies=proxies,
            verify=verify,
            cert=cert,
            http1=http1,
            http2=http2,
            timeout=timeout,
            trust_env=trust_env,
        )

    async def handle_request(
        self, request: "Request"
    ) -> typing.Optional[CLEANUP_WITH_RESPONSE]:
        return await self._process_request_with_middlewares(request=request)

    def send_request(self, request: "Request") -> typing.AsyncGenerator[Response, None]:
        return send_request_with_session(session=self.session, request=request)


def create_default_session(
    app: typing.Optional[typing.Callable[..., typing.Any]],
    base_url: URLTypes,
    cookies: typing.Optional[CookieTypes],
    proxies: typing.Optional[ProxiesTypes],
    verify: typing.Optional[VerifyTypes],
    cert: typing.Optional[CertTypes],
    http1: typing.Optional[bool],
    http2: typing.Optional[bool],
    timeout: typing.Optional[TimeoutTypes],
    trust_env: typing.Optional[bool],
) -> httpx.AsyncClient:
    return httpx.AsyncClient(
        app=app,
        base_url=base_url,
        cookies=first_not_none(cookies, default_configs.DEFAULT_COOKIES),
        proxies=first_not_none(proxies, default_configs.DEFAULT_PROXIES),
        cert=first_not_none(cert, default_configs.DEFAULT_CERTS),
        verify=first_not_none(verify, default_configs.DEFAULT_VERIFY_SSL),
        timeout=first_not_none(timeout, default_configs.REQUEST_TIMEOUT),
        trust_env=first_not_none(trust_env, default_configs.DEFAULT_TRUST_ENV),
        http1=first_not_none(http1, default_configs.HTTP_1),
        http2=first_not_none(http2, default_configs.HTTP_2),
    )


async def send_request_with_session(
    session: httpx.AsyncClient, request: "Request"
) -> typing.AsyncGenerator[Response, None]:
    if request.stream:
        async with session.stream(
            method=request.method,
            url=request.url,
            content=request.content,
            data=request.data,
            files=request.files,
            json=request.json,
            params=request.params,
            headers=request.headers,
            auth=USE_CLIENT_DEFAULT,
            follow_redirects=request.follow_redirects,
        ) as response:
            yield response
    response = await session.request(
        method=request.method,
        url=request.url,
        content=request.content,
        data=request.data,
        files=request.files,
        json=request.json,
        params=request.params,
        headers=request.headers,
        auth=USE_CLIENT_DEFAULT,
        follow_redirects=request.follow_redirects,
    )
    yield response


async def send_request(request: "Request") -> typing.AsyncGenerator[Response, None]:
    async with httpx.AsyncClient(
        cookies=request.cookies,
        proxies=request.proxies,
        cert=request.cert,
        verify=request.verify,
        timeout=request.timeout,
        trust_env=request.trust_env,
        http1=request.http1,
        http2=request.http2,
        base_url=request.base_url,
        app=request.app,
    ) as session:
        if request.stream:
            async with session.stream(
                method=request.method,
                url=request.url,
                content=request.content,
                data=request.data,
                files=request.files,
                json=request.json,
                params=request.params,
                headers=request.headers,
                auth=USE_CLIENT_DEFAULT,
                follow_redirects=request.follow_redirects,
            ) as response:
                yield response
        response = await session.request(
            method=request.method,
            url=request.url,
            content=request.content,
            data=request.data,
            files=request.files,
            json=request.json,
            params=request.params,
            headers=request.headers,
            auth=USE_CLIENT_DEFAULT,
            follow_redirects=request.follow_redirects,
        )
        yield response



import typing
from abc import ABC
from abc import abstractmethod
from enum import Enum
from enum import auto

if typing.TYPE_CHECKING:
    from scrapyio.items import Item


class LoaderState(Enum):
    CREATED = auto()
    OPENED = auto()
    CLOSED = auto()


class BaseLoader(ABC):
    def __init__(self):
        self.state: LoaderState = LoaderState.CREATED

    @abstractmethod
    async def open(self) -> None:
        ...

    @abstractmethod
    async def dump(self, item: "Item") -> None:
        ...

    @abstractmethod
    async def close(self) -> None:
        ...


class JSONLoader(BaseLoader):
    def __init__(self, filename: str):
        super().__init__()
        self.filename = filename
        self.file: typing.Optional[typing.TextIO] = None
        self.first_item: bool = True

    async def open(self) -> None:
        self.state = LoaderState.OPENED
        self.file = open(self.filename, mode="w", encoding="utf-8")
        self.file.write("[")

    async def dump(self, item: "Item") -> None:
        assert self.file, "Loader's `dump` was called before `open` method"
        serialized_item = item.json()
        if not self.first_item:
            self.file.write("," + serialized_item)
        else:
            self.file.write(serialized_item)
            self.first_item = False

    async def close(self) -> None:
        assert self.file, "Loader's `close` was called before `open` method"
        self.file.write("]")
        self.file.close()
        self.state = LoaderState.CLOSED


# TODO: Implement CSVLoader
# TODO: Implement MONGOLoader
# TODO: Implement SQLiteLoader



import typing

T = typing.TypeVar("T")
Y = typing.TypeVar("Y")


def load_module(path: str) -> typing.Any:
    splited_path = path.split(".")
    mod = __import__(".".join(splited_path[:-1]))
    for component in splited_path[1:]:
        mod = getattr(mod, component)
    return mod


def first_not_none(o1: T, o2: Y) -> Y:
    if o1 is not None:
        return typing.cast(Y, o1)
    return o2



import typing
from abc import ABC
from abc import abstractmethod

from .http import Request
from .http import Response
from .types import CLEANUP_WITH_RESPONSE
from .utils import load_module


def build_middlewares_chain() -> typing.List["BaseMiddleWare"]:
    from . import default_configs

    return [
        typing.cast("BaseMiddleWare", load_module(middleware)())
        for middleware in default_configs.MIDDLEWARES
    ]


class BaseMiddleWare(ABC):
    @abstractmethod
    async def process_request(
        self, request: "Request"
    ) -> typing.Union[None, CLEANUP_WITH_RESPONSE]:
        ...

    @abstractmethod
    async def process_response(self, response: Response) -> typing.Union[None, Request]:
        ...



import inspect

import pytest

from scrapyio import Downloader
from scrapyio import IgnoreRequestError
from scrapyio import Request
from scrapyio import Response
from scrapyio import SessionDownloader
from scrapyio.downloader import send_request
from scrapyio.middlewares import BaseMiddleWare


@pytest.mark.anyio
async def test_send_request(app):
    req = Request(url="/", method="GET", app=app, base_url="https://example.am")
    gen = send_request(req)
    response = await gen.__anext__()
    assert response.status_code == 200
    assert response.text == '"Root"'
    with pytest.raises(StopAsyncIteration):
        await gen.__anext__()


@pytest.mark.anyio
async def test_downloader_send_request(app):
    downloader = Downloader()
    req = Request(url="/", method="GET", app=app, base_url="http://example.com")
    cleanup_gen, response = await downloader._process_request_with_middlewares(req)
    assert inspect.isasyncgen(cleanup_gen)
    assert isinstance(response, Response)


@pytest.mark.anyio
async def test_session_downloader_cookies(app):
    downloader = SessionDownloader(app=app, base_url="https://example.com")
    req = Request(url="/cookies", method="GET")
    cleanup_gen, response = await downloader._process_request_with_middlewares(
        request=req
    )
    assert response.status_code == 200
    assert len(downloader.session.cookies) == 1
    cleanup_gen, response = await downloader._process_request_with_middlewares(
        request=req
    )
    assert response.json() == {"test": "test"}


@pytest.mark.anyio
async def test_ignoring_requests():
    class CustomMiddleWare(BaseMiddleWare):
        async def process_response(self, response):
            raise IgnoreRequestError()

        async def process_request(self, request):
            ...

    downloader = Downloader()
    downloader.middlewares.append(CustomMiddleWare())

    resp = await downloader.handle_request(
        Request(url="https://example.com", method="GET")
    )
    assert resp is None


@pytest.mark.anyio
async def test_session_downloader_configs(app):
    downloader = SessionDownloader(
        cookies={"test": "test"}, app=app, base_url="https://example.com"
    )
    gen = await downloader.handle_request(Request(url="/cookies", method="GET"))
    assert gen
    _, response = gen
    assert "test" in response.json()


@pytest.mark.anyio
async def test_session_stream_request(app):
    downloader = SessionDownloader(app=app, base_url="https://example.com")
    gen = await downloader.handle_request(Request(url="/", method="GET", stream=True))
    assert gen
    _, response = gen
    await response.aread()


@pytest.mark.anyio
async def test_stream_request(app):
    downloader = Downloader()
    gen = await downloader.handle_request(
        Request(
            url="/", method="GET", stream=True, app=app, base_url="https://example.com"
        )
    )
    assert gen
    _, response = gen
    await response.aread()



import typing
from abc import ABC
from abc import abstractmethod

from scrapyio.items import Item


class BaseItemMiddleWare(ABC):
    @abstractmethod
    async def process_item(self, item: Item) -> typing.Union[Item, None]:
        ...



import typing
from types import ModuleType

from . import default_configs

LOADED = False

SETTINGS_FILE_NAME = "settings.py"
SETTINGS_FILE_NAME_FOR_IMPORT = "settings"

CONFIGS_TO_LOAD = ("MIDDLEWARES", "REQUEST_TIMEOUT")


def load_settings(path: typing.Optional[str] = None):
    global LOADED
    if LOADED:
        return
    custom_configs: typing.Union[ModuleType, object]
    try:
        custom_configs = __import__(path or SETTINGS_FILE_NAME_FOR_IMPORT)
    except ModuleNotFoundError:
        custom_configs = object()

    for attr in CONFIGS_TO_LOAD:
        value = getattr(custom_configs, attr, None)
        if value:
            setattr(default_configs, attr, value)
    LOADED = True



import typing
from dataclasses import dataclass
from dataclasses import field

from httpx import Response
from httpx._config import Timeout
from httpx._types import AuthTypes
from httpx._types import CertTypes
from httpx._types import CookieTypes
from httpx._types import HeaderTypes
from httpx._types import ProxiesTypes
from httpx._types import QueryParamTypes
from httpx._types import RequestContent
from httpx._types import RequestData
from httpx._types import RequestFiles
from httpx._types import TimeoutTypes
from httpx._types import URLTypes
from httpx._types import VerifyTypes

from . import default_configs


@dataclass
class Request:
    url: str
    method: str
    auth: typing.Optional[AuthTypes] = field(
        default_factory=lambda: default_configs.DEFAULT_AUTH
    )
    params: typing.Optional[QueryParamTypes] = field(
        default_factory=lambda: default_configs.DEFAULT_PARAMS
    )
    headers: typing.Optional[HeaderTypes] = field(
        default_factory=lambda: default_configs.DEFAULT_HEADERS
    )
    cookies: typing.Optional[CookieTypes] = field(
        default_factory=lambda: default_configs.DEFAULT_COOKIES
    )
    verify: VerifyTypes = field(
        default_factory=lambda: default_configs.DEFAULT_VERIFY_SSL
    )
    cert: typing.Optional[CertTypes] = field(
        default_factory=lambda: default_configs.DEFAULT_CERTS
    )
    http1: bool = default_configs.HTTP_1
    http2: bool = default_configs.HTTP_2
    proxies: typing.Optional[ProxiesTypes] = field(
        default_factory=lambda: default_configs.DEFAULT_PROXIES
    )
    follow_redirects: bool = default_configs.FOLLOW_REDIRECTS
    timeout: TimeoutTypes = field(
        default_factory=lambda: Timeout(default_configs.REQUEST_TIMEOUT)
    )
    trust_env: bool = default_configs.DEFAULT_TRUST_ENV
    content: typing.Optional[RequestContent] = None
    data: typing.Optional[RequestData] = None
    files: typing.Optional[RequestFiles] = None
    json: typing.Optional[typing.Any] = None
    stream: bool = default_configs.ENABLE_STREAM_BY_DEFAULT
    app: typing.Optional[typing.Callable[..., typing.Any]] = None
    base_url: URLTypes = ""


async def clean_up_response(response_gen: typing.AsyncGenerator[Response, None]):
    try:
        await response_gen.__anext__()  # Must raise an exception
        assert True, "StopAsyncIteration was expected"  # pragma: no cover
    except StopAsyncIteration:
        ...



from fastapi import FastAPI
from fastapi import Request
from fastapi import Response

app = FastAPI()


@app.get("/")
def root():
    return "Root"


@app.get("/cookies")
def cookies(request: Request, response: Response):
    response.set_cookie(key="test", value="test")
    return request.cookies


@app.get("/headers")
def headers(request: Request):
    return request.headers


@app.get("/for_moving")
def moving():
    return 200


@app.get("/best_scraping_library")
def best_scraping_library():
    return "scrapyio"



from .downloader import Downloader
from .downloader import SessionDownloader
from .exceptions import IgnoreRequestError
from .spider import BaseSpider
from .spider import Request
from .spider import Response



import asyncio
import typing
from abc import ABC
from abc import abstractmethod

from pydantic import BaseModel

from . import default_configs
from .exceptions import IgnoreItemError
from .item_loaders import BaseLoader
from .item_loaders import LoaderState
from .types import ITEM_ADDED_CALLBACK_TYPE
from .types import ITEM_IGNORING_CALLBACK_TYPE
from .utils import load_module

if typing.TYPE_CHECKING:
    from .item_middlewares import BaseItemMiddleWare


def build_items_middlewares_chain() -> typing.Sequence["BaseItemMiddleWare"]:
    return [
        typing.cast("BaseItemMiddleWare", load_module(middleware)())
        for middleware in default_configs.ITEM_MIDDLEWARES
    ]


def orjson_dumps_wrapper(*args, **kwargs) -> str:
    import orjson

    return orjson.dumps(*args, **kwargs).decode(encoding="utf-8")


class BaseItem(BaseModel):
    class Config:
        try:
            import orjson

            json_dumps = orjson_dumps_wrapper
            json_loads = orjson.loads
        except ImportError:
            ...


class Item(BaseItem):
    ...


class BaseItemsManager(ABC):
    def __init__(
        self,
        ignoring_callback: typing.Optional[ITEM_IGNORING_CALLBACK_TYPE] = None,
        success_callback: typing.Optional[ITEM_ADDED_CALLBACK_TYPE] = None,
        loader: typing.Optional[BaseLoader] = None,
    ):
        self.middlewares = build_items_middlewares_chain()
        self.ignoring_callback = ignoring_callback
        self.success_callback = success_callback
        self.loader = loader

    async def _send_single_item_via_middlewares(
        self, item: Item
    ) -> typing.Optional[Item]:
        for middleware in self.middlewares:
            try:
                await middleware.process_item(item=item)
            except IgnoreItemError:
                if self.ignoring_callback:
                    await self.ignoring_callback(item, middleware)
                return None
        if self.success_callback:
            await self.success_callback(item)

        return item

    async def _send_items_via_middlewares(
        self, items: typing.Sequence[Item]
    ) -> typing.Sequence[Item]:
        tasks = [
            asyncio.create_task(self._send_single_item_via_middlewares(item))
            for item in items
        ]

        filtered_items = [
            added_item for added_item in await asyncio.gather(*tasks) if added_item
        ]

        if self.loader:
            if self.loader.state == LoaderState.CREATED:
                await self.loader.open()
            loading_tasks = [
                asyncio.create_task(self.loader.dump(item_to_load))
                for item_to_load in filtered_items
            ]
            await asyncio.gather(
                *loading_tasks,
            )

        return typing.cast(typing.List[Item], filtered_items)

    @abstractmethod
    async def process_items(self, items: typing.Sequence[Item]) -> None:
        ...


class ItemManager(BaseItemsManager):
    async def process_items(self, items: typing.Sequence[Item]) -> None:
        await self._send_items_via_middlewares(items=items)
        return None



from scrapyio.downloader import BaseDownloader
from scrapyio.middlewares import BaseMiddleWare
from scrapyio.utils import load_module


def test_object_loading():
    assert BaseMiddleWare == load_module("scrapyio.middlewares.BaseMiddleWare")
    assert BaseDownloader == load_module("scrapyio.downloader.BaseDownloader")
    assert load_module == load_module("scrapyio.utils.load_module")



import typing

from httpx._types import AuthTypes
from httpx._types import CertTypes
from httpx._types import CookieTypes
from httpx._types import HeaderTypes
from httpx._types import ProxiesTypes
from httpx._types import QueryParamTypes
from httpx._types import VerifyTypes

ITEM_MIDDLEWARES: typing.List[str] = []

# Middlewares
#   path to the middlewares
#   example: 'scrapyio.middlewares.BaseMiddleWare'
MIDDLEWARES: typing.List[str] = []

# Timeout for httpx request
REQUEST_TIMEOUT: int = 5

# Default headers for httpx request
DEFAULT_HEADERS: typing.Optional[HeaderTypes] = None

# Default cookies for httpx request
DEFAULT_COOKIES: typing.Optional[CookieTypes] = {}

# Default query parameters for httpx request
DEFAULT_PARAMS: typing.Optional[QueryParamTypes] = None

# Default auth for httpx request
DEFAULT_AUTH: typing.Optional[AuthTypes] = None

# Default for SSL verify mode
DEFAULT_VERIFY_SSL: VerifyTypes = True

# Default certs for httpx request
DEFAULT_CERTS: typing.Optional[CertTypes] = None

# Default HTTP version
HTTP_1: bool = True
HTTP_2: bool = False

# Default HTTP proxies
DEFAULT_PROXIES: typing.Optional[ProxiesTypes] = None

# Follow redirects for HTTP request
FOLLOW_REDIRECTS: bool = False

# Trust env for httpx request
DEFAULT_TRUST_ENV: bool = False

# Enable stream by default
ENABLE_STREAM_BY_DEFAULT: bool = False



from pathlib import Path
from tempfile import NamedTemporaryFile

from scrapyio import default_configs
from scrapyio import settings
from scrapyio.settings import load_settings


def test_load_settings(monkeypatch):
    monkeypatch.setattr(settings, "LOADED", False)
    saved_defaults = open(default_configs.__file__).read()
    with NamedTemporaryFile(mode="w+", suffix=".py", dir=".") as file:
        path = Path(file.name)
        import_name = path.name[:-3]  # without .py suffix
        monkeypatch.syspath_prepend(path=path.parent)
        file.write(saved_defaults)
        file.flush()
        module = __import__(import_name)
        module.REQUEST_TIMEOUT = 10
        module.MIDDLEWARES = ["NOT A PATH"]
        load_settings(import_name)
        assert default_configs.REQUEST_TIMEOUT == 10
        assert default_configs.MIDDLEWARES == ["NOT A PATH"]
        module.REQUEST_TIMEOUT = 5
        load_settings(import_name)
        assert default_configs.REQUEST_TIMEOUT != 5


def test_load_settings_without_file(monkeypatch):
    assert not settings.LOADED
    UNEXISTING_FILE_NAME = "NOTEXISTS"
    UNEXISTING_FILE_NAME_FOR_IMPORT = "NOTEXISTS.py"
    monkeypatch.setattr(settings, "SETTINGS_FILE_NAME", UNEXISTING_FILE_NAME)
    monkeypatch.setattr(
        settings, "SETTINGS_FILE_NAME_FOR_IMPORT", UNEXISTING_FILE_NAME_FOR_IMPORT
    )
    load_settings()



import builtins
import importlib
import json
import os
import tempfile
import types
import typing

import pytest

from scrapyio import items
from scrapyio.item_loaders import JSONLoader
from scrapyio.items import Item
from scrapyio.items import ItemManager


@pytest.mark.anyio
async def test_loader_set_up():
    class MyItem(Item):
        library_name: str

    filename = tempfile.mktemp()
    loader = JSONLoader(filename=filename)
    manager = ItemManager(loader=loader)
    try:
        await manager.process_items([MyItem(library_name="scrapyio")])
    finally:
        await manager.loader.close()
        os.remove(filename)


@pytest.mark.anyio
async def test_json_loader_without_orjson(monkeypatch):
    old_import = builtins.__import__

    def fake_import(name: str, *args, **kwargs) -> types.ModuleType:
        if name == "orjson":
            raise ImportError()
        return typing.cast(types.ModuleType, old_import(name, *args, **kwargs))

    monkeypatch.setattr(builtins, "__import__", fake_import)
    importlib.reload(items)
    monkeypatch.setattr(builtins, "__import__", old_import)

    class MyItem(Item):
        library_name: str

    filename = tempfile.mktemp()
    loader = JSONLoader(filename=filename)
    manager = ItemManager(loader=loader)
    try:
        await manager.process_items([MyItem(library_name="scrapyio")])
    finally:
        await manager.loader.close()
        with open(filename, mode="r", encoding="utf-8") as f:
            dumped = json.loads(f.read())
            assert len(dumped) == 1
            assert dumped[0]["library_name"] == "scrapyio"
        os.remove(filename)


@pytest.mark.anyio
async def test_json_loader():
    pytest.importorskip("orjson")

    class MyItem(Item):
        library_name: str

    filename = tempfile.mktemp()
    loader = JSONLoader(filename=filename)
    manager = ItemManager(loader=loader)
    try:
        await manager.process_items(
            [MyItem(library_name="scrapyio"), MyItem(library_name="scrapyio")]
        )
    finally:
        await manager.loader.close()
        with open(filename, mode="r", encoding="utf-8") as f:
            content = f.read()
            dumped = json.loads(content)
            assert len(dumped) == 2
            assert dumped[0]["library_name"] == "scrapyio"
        os.remove(filename)



import typing
from abc import ABC
from abc import abstractmethod

from httpx._client import Response

from .http import Request
from .items import Item


class BaseSpider(ABC):
    start_requests: typing.ClassVar[typing.List[Request]] = []

    def __init__(self):
        self.requests = self.start_requests  # pragma: no cover
        self.items: typing.List[Item] = []  # pragma: no cover

    @abstractmethod
    async def parse(
        self, response: Response
    ) -> typing.AsyncGenerator[typing.Union[Request, Item, None], None]:
        if False:
            yield Request()



from pathlib import Path

import click

from scrapyio import default_configs
from scrapyio.settings import SETTINGS_FILE_NAME


@click.group()
def main():
    ...


@main.command()
@click.argument("name")
def new(name):
    path = Path.cwd()
    dir_path = path / name
    dir_path.mkdir()
    settings_file = dir_path / SETTINGS_FILE_NAME
    settings_file.touch()
    settings_file.write_text(open(default_configs.__file__).read())


@main.command()
def run(test):
    print("main")



import typing

import pytest

from scrapyio.exceptions import IgnoreItemError
from scrapyio.item_middlewares import BaseItemMiddleWare
from scrapyio.items import Item
from scrapyio.items import ItemManager


@pytest.mark.anyio
async def test_item_manager():
    class MyItem(Item):
        library_name: str

    items: typing.List[Item] = [MyItem(library_name="scrapyio")]
    manager = ItemManager()
    await manager.process_items(items)


@pytest.mark.anyio
async def test_item_manager_callbacks():
    class MyItem(Item):
        library_name: str

    class MyMiddleWare(BaseItemMiddleWare):
        async def process_item(self, item: Item) -> typing.Union[Item, None]:
            raise IgnoreItemError()

    async def my_success_callback(item: MyItem):
        SUCCESS.append(item.library_name)

    async def my_failed_loading_callback(item: MyItem, middleware: MyMiddleWare):
        FAILED.append((item.library_name, middleware))

    SUCCESS: typing.List[str] = []
    FAILED: typing.List[typing.Tuple[str, MyMiddleWare]] = []

    items: typing.List[Item] = [MyItem(library_name="scrapyio")]
    manager = ItemManager(success_callback=my_success_callback)
    await manager.process_items(items)
    assert SUCCESS == ["scrapyio"]
    manager.middlewares.append(MyMiddleWare())
    manager.ignoring_callback = my_failed_loading_callback
    await manager.process_items(items)
    assert len(SUCCESS) == 1
    assert len(FAILED) == 1
    assert FAILED[0][0] == "scrapyio"
    assert FAILED[0][1] == manager.middlewares[-1]


@pytest.mark.anyio
async def test_items_filtering():
    class MyItem(Item):
        library_name: str

    class MyMiddleWare(BaseItemMiddleWare):
        async def process_item(self, item: Item) -> typing.Union[Item, None]:
            item = typing.cast(MyItem, item)
            if item.library_name != "scrapyio":
                raise IgnoreItemError()
            return item

    async def succes_callback(item: MyItem):
        SUCCESS.append(item)

    SUCCESS: typing.List[MyItem] = []

    manager = ItemManager(success_callback=succes_callback)
    manager.middlewares.append(MyMiddleWare())
    items: typing.List[Item] = [
        MyItem(library_name="scrapyio"),
        MyItem(library_name="scrapy"),
    ]
    await manager.process_items(items=items)
    assert len(SUCCESS) == 1
    assert SUCCESS[0].library_name == "scrapyio"
    item_added = await manager._send_single_item_via_middlewares(
        MyItem(library_name="scrapy")
    )
    assert not item_added



from abc import ABC

import pytest

from scrapyio import Request
from scrapyio.downloader import Downloader
from scrapyio.middlewares import BaseMiddleWare


@pytest.mark.anyio
async def test_request_middleware(app):
    class CustomMiddleWare(BaseMiddleWare, ABC):
        async def process_response(self, response):
            ...

        async def process_request(self, request):
            request.headers["test"] = "test"

    downloader = Downloader()
    downloader.middlewares.append(CustomMiddleWare())
    gen = await downloader.handle_request(
        request=Request(
            url="/headers",
            app=app,
            headers={"test": "test"},
            base_url="http://example.com",
            method="GET",
        )
    )
    assert gen
    _, response = gen[0], gen[1]
    assert response.status_code == 200
    assert "test" in response.json()


@pytest.mark.anyio
async def test_response_middleware(app):
    class CustomMiddleWare(BaseMiddleWare):
        async def process_response(self, response):
            response.headers["test"] = "test"

        async def process_request(self, request):
            ...

    downloader = Downloader()
    downloader.middlewares.append(CustomMiddleWare())
    gen = await downloader.handle_request(
        request=Request(
            url="/",
            app=app,
            headers={"test": "test"},
            base_url="http://example.com",
            method="GET",
        )
    )
    assert gen
    _, response = gen[0], gen[1]
    assert response.status_code == 200
    assert "test" in response.headers


@pytest.mark.anyio
async def test_explicit_response_return_middleware(app):
    class CustomMiddleWare(BaseMiddleWare):
        async def process_response(self, response):
            ...

        async def process_request(self, request):
            async def do_request():
                import httpx

                async with httpx.AsyncClient(app=app) as session:
                    response = await session.get(url="https://replaced.com")
                    yield response

            gen = do_request()
            response = await gen.__anext__()
            return gen, response

    downloader = Downloader()
    downloader.middlewares.append(CustomMiddleWare())
    gen = await downloader.handle_request(
        request=Request(url="http://example.com", method="GET")
    )
    assert gen
    _, response = gen[0], gen[1]
    assert response.status_code == 200
    assert response.request.url.scheme == "https"
    assert response.request.url.host == "replaced.com"


@pytest.mark.anyio
async def test_explicit_request_return_middleware(app):
    class CustomMiddleWare(BaseMiddleWare):
        USED = True

        async def process_response(self, response):
            if CustomMiddleWare.USED:  # to prevent recursion
                CustomMiddleWare.USED = False
                return Request(
                    url="/for_moving",
                    app=app,
                    method="GET",
                    base_url="https://example.com",
                )

        async def process_request(self, request):
            ...

    downloader = Downloader()
    downloader.middlewares.append(CustomMiddleWare())
    gen = await downloader.handle_request(
        request=Request(url="http://example.com/", method="GET")
    )
    assert gen
    _, response = gen[0], gen[1]
    assert response.status_code == 200
    assert response.request.url.path == "/for_moving"  # moved


@pytest.mark.anyio
async def test_invalid_middlewares_returns(app):
    class CustomMiddleWare(BaseMiddleWare):
        async def process_response(self, response):
            return 5

        async def process_request(self, request):
            ...

    downloader = Downloader()
    downloader.middlewares.append(CustomMiddleWare())
    with pytest.raises(TypeError, match="Response processing middleware must return.*"):
        await downloader.handle_request(
            request=Request(url="http://example.com", method="GET")
        )

    class CustomMiddleWare(BaseMiddleWare):
        async def process_response(self, response):
            raise NotImplementedError

        async def process_request(self, request):
            return 5

    downloader.middlewares[-1] = CustomMiddleWare()
    with pytest.raises(TypeError, match="Request processing middleware must return"):
        await downloader.handle_request(
            request=Request(url="http://example.com", method="GET")
        )



import asyncio
import inspect
import typing
from warnings import warn

from scrapyio import Request
from scrapyio.downloader import BaseDownloader
from scrapyio.downloader import Downloader
from scrapyio.http import clean_up_response
from scrapyio.item_loaders import BaseLoader
from scrapyio.items import ItemManager
from scrapyio.settings import load_settings
from scrapyio.spider import BaseSpider
from scrapyio.spider import Item
from scrapyio.types import CLEANUP_WITH_RESPONSE

from .utils import first_not_none


class Engine:
    downloader_class: typing.ClassVar[typing.Type[BaseDownloader]] = Downloader
    items_manager_class: typing.ClassVar[
        typing.Optional[typing.Type[ItemManager]]
    ] = None
    loader_class: typing.ClassVar[typing.Optional[typing.Type[BaseLoader]]] = None

    def __init__(
        self,
        spider_class: typing.Type[BaseSpider],
        downloader_class: typing.Optional[typing.Type[BaseDownloader]] = None,
        items_manager_class: typing.Optional[typing.Type[ItemManager]] = None,
        enable_settings: bool = True,
    ):
        if enable_settings:
            load_settings()  # pragma: no cover
        downloader_class = first_not_none(downloader_class, self.downloader_class)
        items_manager_class = first_not_none(
            items_manager_class, self.items_manager_class
        )

        self.downloader: BaseDownloader = downloader_class()
        self.spider = spider_class()
        self.items_manager: typing.Optional[ItemManager] = (
            items_manager_class() if items_manager_class else None
        )

        if self.items_manager is None:
            warn(
                "Because no `items_manager` was specified, all items"
                " yielded by the 'parse' method will be ignored.",
                RuntimeWarning,
            )

    async def _send_single_request_to_downloader(
        self, request: Request
    ) -> typing.Optional[CLEANUP_WITH_RESPONSE]:
        return await self.downloader.handle_request(request=request)

    async def _send_all_requests_to_downloader(
        self,
    ) -> typing.List[CLEANUP_WITH_RESPONSE]:
        request_tasks: typing.List[typing.Awaitable] = []
        for request in self.spider.requests:
            request_tasks.append(
                asyncio.create_task(
                    self._send_single_request_to_downloader(request=request)
                )
            )
        self.spider.requests.clear()
        responses: typing.Iterable[
            typing.Optional[CLEANUP_WITH_RESPONSE]
        ] = await asyncio.gather(
            *request_tasks
        )  # type: ignore
        return [response for response in responses if response]

    async def _handle_single_response(
        self, response_and_generator: CLEANUP_WITH_RESPONSE
    ) -> None:
        if not inspect.isasyncgenfunction(self.spider.parse):
            raise TypeError(
                "Spider's `parse` must be an asynchronous generator function"
            )
        clean_up_generator, response = response_and_generator
        try:
            gen = self.spider.parse(response=response)
            async for yielded_value in gen:
                if isinstance(yielded_value, Request):
                    self.spider.requests.append(yielded_value)
                elif isinstance(yielded_value, Item):
                    self.spider.items.append(yielded_value)  # pragma: no cover
                elif yielded_value is None:
                    ...
                else:
                    raise TypeError(
                        "Invalid type yielded, expected `Request` or `Item` got `%s`"
                        % yielded_value.__class__.__name__
                    )
        finally:
            await clean_up_response(clean_up_generator)

    async def _handle_responses(
        self, responses: typing.List[CLEANUP_WITH_RESPONSE]
    ) -> None:
        tasks = [
            asyncio.create_task(self._handle_single_response(response))
            for response in responses
        ]
        await asyncio.gather(*tasks)

    async def run_once(self):
        responses = await self._send_all_requests_to_downloader()
        await self._handle_responses(responses=responses)
        if self.items_manager:
            await self.items_manager.process_items(self.spider.items)

        self.spider.items.clear()

    async def run(self):
        while self.spider.requests:
            await self.run_once()



import typing

from httpx._client import Response

if typing.TYPE_CHECKING:
    from .item_middlewares import BaseItemMiddleWare
    from .items import Item

RESPONSE_GENERATOR = typing.AsyncGenerator[Response, None]
CLEANUP_WITH_RESPONSE = typing.Tuple[RESPONSE_GENERATOR, Response]

ITEM_IGNORING_CALLBACK_TYPE = typing.Callable[
    ["Item", "BaseItemMiddleWare"], typing.Coroutine[typing.Any, typing.Any, typing.Any]
]

ITEM_ADDED_CALLBACK_TYPE = typing.Callable[
    ["Item"], typing.Coroutine[typing.Any, typing.Any, typing.Any]
]



import typing
from warnings import catch_warnings

import pytest

from scrapyio import Request
from scrapyio import Response
from scrapyio.downloader import Downloader
from scrapyio.downloader import SessionDownloader
from scrapyio.engines import Engine
from scrapyio.items import Item
from scrapyio.items import ItemManager
from scrapyio.spider import BaseSpider


@pytest.mark.anyio
async def test_engine_request_handling(app):
    class Spider(BaseSpider):
        start_requests = [
            Request(url="/", base_url="https://example.com", method="GET", app=app)
        ]

        async def parse(self, response):
            raise NotImplementedError

    engine = Engine(spider_class=Spider, enable_settings=False)
    response_with_cleanup = await engine._send_single_request_to_downloader(
        Request(url="/", base_url="https://example.com", method="GET", app=app)
    )
    assert response_with_cleanup
    _, response = response_with_cleanup
    assert response.status_code == 200
    assert len(engine.spider.requests) == 1
    responses_with_clean_ups = await engine._send_all_requests_to_downloader()
    assert len(responses_with_clean_ups) == 1
    _, response = responses_with_clean_ups[0]
    assert response.status_code == 200


@pytest.mark.anyio
async def test_engine_response_handling(app):
    test_list = []

    class Spider(BaseSpider):
        start_requests = [
            Request(url="/", base_url="https://example.com", method="GET", app=app)
        ]

        async def parse(self, response):
            test_list.append(response.text)
            yield None

    engine = Engine(spider_class=Spider, enable_settings=False)
    responses = await engine._send_all_requests_to_downloader()
    await engine._handle_responses(responses=responses)
    assert len(test_list) == 1


@pytest.mark.anyio
async def test_engine_sync_parser_error(app):
    class Spider(BaseSpider):
        start_requests = [
            Request(url="/", base_url="https://example.com", method="GET", app=app)
        ]

        def parse(self, response):
            raise NotImplementedError

    engine = Engine(spider_class=Spider, enable_settings=False)
    with pytest.raises(TypeError, match="Spider's `parse`.*"):
        await engine.run()


@pytest.mark.anyio
async def test_engine_parser_invalid_yield_value(app):
    class Spider(BaseSpider):
        start_requests = [
            Request(url="/", base_url="https://example.com", method="GET", app=app)
        ]

        async def parse(self, response):
            yield 5

    engine = Engine(spider_class=Spider, enable_settings=False)
    with pytest.raises(TypeError, match="Invalid type yielded,.*"):
        await engine.run()


@pytest.mark.anyio
async def test_engine_parser_yield_request(app):
    class Spider(BaseSpider):
        start_requests = [
            Request(url="/", base_url="https://example.com", method="GET", app=app)
        ]

        async def parse(self, response):
            yield Request(
                url="/", base_url="https://example.com", method="GET", app=app
            )

    engine = Engine(spider_class=Spider, enable_settings=False)
    responses = await engine._send_all_requests_to_downloader()
    assert engine.spider.requests == []
    await engine._handle_responses(responses)
    assert len(engine.spider.requests) == 1


def test_engine_configs():
    class FakeSpider:
        ...

    class MyEngine(Engine):
        downloader_class = SessionDownloader

    session_engine = MyEngine(spider_class=FakeSpider, enable_settings=False)
    assert isinstance(session_engine.downloader, SessionDownloader)
    engine = MyEngine(
        spider_class=FakeSpider, enable_settings=False, downloader_class=Downloader
    )
    assert isinstance(engine.downloader, Downloader)


@pytest.mark.anyio
async def test_engine_item_processing(app):
    class MyItem(Item):
        best_scraping_library: str

    class Spider(BaseSpider):
        start_requests = [
            Request(
                url="/best_scraping_library",
                method="GET",
                base_url="https://example.com",
                app=app,
            )
        ]

        async def parse(
            self, response: Response
        ) -> typing.AsyncGenerator[typing.Union[Request, Item], None]:
            yield MyItem(best_scraping_library=response.text)
            yield MyItem(best_scraping_library=response.text)
            yield Request(
                url="/best_scraping_library",
                method="GET",
                base_url="https://example.com",
                app=app,
            )

    engine = Engine(spider_class=Spider, enable_settings=False)
    await engine.run_once()
    assert len(engine.spider.requests) == 1
    assert engine.spider.items == []


@pytest.mark.filterwarnings("once::RuntimeWarning")
@pytest.mark.anyio
def test_engine_without_items_manager_warning():
    with catch_warnings(record=True) as w:
        Engine(spider_class=lambda: 0, enable_settings=False)
        assert w
        assert len(w) == 1
        (warning,) = w
        assert warning.category == RuntimeWarning


@pytest.mark.anyio
async def test_engine_with_item_manager(app):
    class Spider(BaseSpider):
        start_requests = [
            Request(
                url="/best_scraping_library",
                method="GET",
                base_url="https://example.com",
                app=app,
            )
        ]

        async def parse(
            self, response: Response
        ) -> typing.AsyncGenerator[typing.Union[Request, Item, None], None]:
            yield None

    engine = Engine(
        spider_class=Spider, items_manager_class=ItemManager, enable_settings=False
    )
    await engine.run_once()
    assert engine.spider.requests == []
